package layer

import (
	"context"
	"fmt"
	"sort"
	"strconv"
	"strings"

	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/api"
	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/api/data"
	apierr "git.frostfs.info/TrueCloudLab/frostfs-s3-gw/api/errors"
	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/api/layer/frostfs"
	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/api/middleware"
	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/internal/logs"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/container"
	"go.uber.org/zap"
)

const (
	attributeLocationConstraint = ".s3-location-constraint"
	AttributeLockEnabled        = "LockEnabled"
)

func (n *Layer) containerInfo(ctx context.Context, prm frostfs.PrmContainer) (*data.BucketInfo, error) {
	var (
		err error
		res *container.Container
		log = n.reqLogger(ctx).With(zap.Stringer("cid", prm.ContainerID))

		info = &data.BucketInfo{
			CID:  prm.ContainerID,
			Name: prm.ContainerID.EncodeToString(),
		}

		reqInfo = middleware.GetReqInfo(ctx)
	)

	res, err = n.frostFS.Container(ctx, prm)
	if err != nil {
		if client.IsErrContainerNotFound(err) {
			return nil, fmt.Errorf("%w: %s", apierr.GetAPIError(apierr.ErrNoSuchBucket), err.Error())
		}
		return nil, fmt.Errorf("get frostfs container: %w", err)
	}

	cnr := *res

	info.Owner = cnr.Owner()
	if domain := container.ReadDomain(cnr); domain.Name() != "" {
		info.Name = domain.Name()
		info.Zone = domain.Zone()
	}
	info.Created = container.CreatedAt(cnr)
	info.LocationConstraint = cnr.Attribute(attributeLocationConstraint)
	info.HomomorphicHashDisabled = container.IsHomomorphicHashingDisabled(cnr)
	info.PlacementPolicy = cnr.PlacementPolicy()

	attrLockEnabled := cnr.Attribute(AttributeLockEnabled)
	if len(attrLockEnabled) > 0 {
		info.ObjectLockEnabled, err = strconv.ParseBool(attrLockEnabled)
		if err != nil {
			log.Error(logs.CouldNotParseContainerObjectLockEnabledAttribute,
				zap.String("lock_enabled", attrLockEnabled),
				zap.Error(err),
				logs.TagField(logs.TagDatapath),
			)
		}
	}

	zone := n.features.FormContainerZone(reqInfo.Namespace)
	if zone != info.Zone {
		return nil, fmt.Errorf("ns '%s' and zone '%s' are mismatched for container '%s'", zone, info.Zone, prm.ContainerID)
	}

	n.cache.PutBucket(info)

	return info, nil
}

func (n *Layer) containerList(ctx context.Context, listParams ListBucketsParams) ([]*data.BucketInfo, error) {
	stoken := n.SessionTokenForRead(ctx)

	prm := frostfs.PrmUserContainers{
		UserID:       n.BearerOwner(ctx),
		SessionToken: stoken,
	}

	res, err := n.frostFS.UserContainers(ctx, prm)
	if err != nil {
		n.reqLogger(ctx).Error(logs.CouldNotListUserContainers, zap.Error(err), logs.TagField(logs.TagExternalStorage))
		return nil, err
	}

	list := make([]*data.BucketInfo, 0, len(res))
	for i := range res {
		getPrm := frostfs.PrmContainer{
			ContainerID:  res[i],
			SessionToken: stoken,
		}
		info, err := n.containerInfo(ctx, getPrm)
		if err != nil {
			n.reqLogger(ctx).Error(logs.CouldNotFetchContainerInfo, zap.Error(err), logs.TagField(logs.TagExternalStorage))
			continue
		}

		if shouldSkipBucket(info, listParams) {
			continue
		}

		list = append(list, info)
	}

	sort.Slice(list, func(i, j int) bool {
		return list[i].Name < list[j].Name
	})

	for i, info := range list {
		if listParams.ContinuationToken != "" && info.Name != listParams.ContinuationToken {
			continue
		}
		return list[i:], nil
	}

	return nil, nil
}

func shouldSkipBucket(info *data.BucketInfo, prm ListBucketsParams) bool {
	if !strings.HasPrefix(info.Name, prm.Prefix) ||
		(prm.BucketRegion != "" && info.LocationConstraint != prm.BucketRegion) {
		return true
	}

	return false
}

func (n *Layer) createContainer(ctx context.Context, p *CreateBucketParams) (*data.BucketInfo, error) {
	if p.LocationConstraint == "" {
		p.LocationConstraint = api.DefaultLocationConstraint // s3tests_boto3.functional.test_s3:test_bucket_get_location
	}

	zone := n.features.FormContainerZone(p.Namespace)

	bktInfo := &data.BucketInfo{
		Name:               p.Name,
		Zone:               zone,
		Owner:              n.BearerOwner(ctx),
		Created:            TimeNow(ctx),
		LocationConstraint: p.LocationConstraint,
		ObjectLockEnabled:  p.ObjectLockEnabled,
	}

	attributes := [][2]string{
		{attributeLocationConstraint, p.LocationConstraint},
	}

	if p.ObjectLockEnabled {
		attributes = append(attributes, [2]string{
			AttributeLockEnabled, "true",
		})
	}

	res, err := n.frostFS.CreateContainer(ctx, frostfs.PrmContainerCreate{
		Creator:              bktInfo.Owner,
		Policy:               p.Policy,
		Name:                 p.Name,
		Zone:                 zone,
		SessionToken:         p.SessionContainerCreation,
		CreationTime:         bktInfo.Created,
		AdditionalAttributes: attributes,
	})
	if err != nil {
		return nil, fmt.Errorf("create container: %w", err)
	}

	bktInfo.CID = res.ContainerID
	bktInfo.HomomorphicHashDisabled = res.HomomorphicHashDisabled
	bktInfo.PlacementPolicy = p.Policy

	n.cache.PutBucket(bktInfo)

	return bktInfo, nil
}
